# Copyright (c) 2007 Michael "Z" Goddard
# This software is distributed under the MIT license.

"""
Contains functions and classes that analyze config.py and break it up into
sections of variables allowing the webfront to alter it in an object oriented
fashion.
"""

import os
import sys
import pickle
import types

from Pyblosxom.tools import run_callback

import fileops

def findfile():
    for path in [p for p in sys.path if os.path.isdir(p)]:
        for file in os.listdir(path):
            if file == "config.py":
                return os.path.join(path, file)
    raise StandardError, "Cannot find 'config.py'."

def getconfigfile(mode='r'):
    return file(findfile(), mode)

def getregisteredfile(mode='r'):
    return fileops.getTemporaryFile("registered", mode)

def iscomment(line):
    return line.startswith('#')

def ishorizontalrule(line):
    content = line.strip('# \n\r')
    return iscomment(line) and \
           content.count('=') == len(content) > 0

def isblank(line):
    return len(line.strip().strip(os.linesep)) == 0


class ConfigEditor:
    def __init__(self, fileobj=None):
        self.pretext = ""
        self.sections = []
        if fileobj != None and hasattr(fileobj, 'readlines'):
            self.file = fileobj
        else:
            self.file = getconfigfile()
        self.parse(self.file.readlines())
        self.file.seek(0)
        self.mergeRegistered()
    
    def __del__(self):
        self.saveRegistered()
    
    def __str__(self):
        retval = ""
        retval += self.pretext
        retval += 3*os.linesep
        for s in self.sections:
            retval += str(s)
        return retval
    
    def __iter__(self):
        yield self.sections.__iter__()
    
    def parse(self, lines):
        self.pretext = ""
        self.sections = []
        hr = 0
        while hr < 2:
            if ishorizontalrule(lines[0]):
                hr += 1
            self.pretext += lines.pop(0)
        while not ConfigSection.isSection(lines) and len(lines) != 0:
            self.pretext += lines.pop(0)
        while len(lines) != 0:
            if ConfigSection.isSection(lines):
                section = ConfigSection.parse(lines)
                if section != None:
                    self.sections.append(section)
            else:
                lines.pop(0)
    
    def save(self):
        getconfigfile(mode='w').write(str(self))
    
    def mergeRegistered(self):
        configs = []
        run_callback('webfront_config',
            {'configurations': configs})
        self.loadRegistered()
        cls = ConfigEditor
        for conf in configs:
            if isinstance(conf, types.FunctionType):
                conf()
            elif not cls.registered.has_key(conf[1]) or \
               cls.registered[conf[1]] < conf[2]:
                conf[0]()
                ConfigEditor.registerName(conf[1], conf[2])
        for name, section in cls.register.items():
            if self.hasSection(name):
                self.getSection(name).extend(section)
            else:
                self.addSection(section)
    
    def hasVariable(self, name):
        return bool(self.getVariable(name))
    
    def getVariable(self, name):
        for section in self.sections:
            if section.hasVariable(name):
                return section.getVariable(name)
        return None
    
    def getSection(self, name):
        for section in self.sections:
            if section.name == name:
                return section
        return None
    
    def hasSection(self, name):
        return bool(self.getSection(name))
    
    def addSection(self, section):
        self.sections.append(section)
    
    def delSection(self, name):
        if isinstance(name, ConfigSection):
            self.sections.remove(name)
        for section in self.sections:
            if section.name == name:
                self.sections.remove(section)
                return
    
    def loadRegistered(cls):
        if not hasattr(cls, "register"):
            registerfile = getregisteredfile()
            try:
                cls.register = pickle.load(registerfile)
                cls.registered = pickle.load(registerfile)
            except:
                cls.register = {}
                cls.registered = {}
    
    def saveRegistered(cls):
        if hasattr(cls, "register"):
            regfile = getregisteredfile(mode='w')
            pickle.dump(cls.register, regfile)
            pickle.dump(cls.registered, regfile)
    
    def registerName(cls, name, version):
        cls.registered[name] = version
    
    def registerSection(cls, section):
        cls.loadRegistered()
        cls.register[section.name] = section
        cls.saveRegistered()
    
    def getRegisteredSection(cls, name):
        cls.loadRegistered()
        return cls.register[name]
    
    def hasRegisteredSection(cls, name):
        return bool(cls.getRegisteredSection(name))
    
    loadRegistered = classmethod(loadRegistered)
    saveRegistered = classmethod(saveRegistered)
    registerName = classmethod(registerName)
    registerSection = classmethod(registerSection)
    getRegisteredSection = classmethod(getRegisteredSection)
    hasRegisteredSection = classmethod(hasRegisteredSection)
    

class ConfigSection:
    def __init__(self, name, variables=[]):
        self.name = name
        self.variables = variables
    
    def __repr__(self):
        clsname = self.__class__
        clsname = "%s.%s" % (clsname.__module__, clsname.__name__)
        return "<%s name='%s'>" % (clsname, self.name)
    
    def __str__(self):
        retval = "# " + self.name + os.linesep
        retval += "# " + len(self.name)*'=' + 2*os.linesep
        for v in self.variables:
            retval += str(v)
        retval += 2*os.linesep
        return retval
    
    def __iter__(self):
        yield self.variables.__iter__()
    
    def extend(self, section):
        for variable in section.variables:
            if self.hasVariable(variable.name):
                var = self.getVariable(variable.name)
                var.doc = variable.doc
            else:
                self.variables.append(variable)
    
    def hasVariable(self, name):
        return bool(self.getVariable(name))
    
    def getVariable(self, name):
        for variable in self.variables:
            if variable.name == name:
                return variable
        return None
    
    def addVariable(self, variable):
        self.variables.append(variable)
    
    def delVariable(self, name):
        if isinstance(name, ConfigVariable):
            self.variables.remove(name)
        for variable in self.variables:
            if variable.name == name:
                self.variables.remove(variable)
                return
    
    def isSection(lines):
        if len(lines) < 2:
            return False
        emptybefore = isblank(lines[0])
        for l, l2 in zip(lines, lines[1:]):
            if emptybefore and isblank(l):
                continue
            elif isblank(l):
                break
            emptybefore = False
            if iscomment(l) and \
               l2 != None and \
               ishorizontalrule(l2):
                return True
        return False
    
    def parse(cls, lines):
        while isblank(lines[0]):
            lines.pop(0)
        while not (iscomment(lines[0]) and \
                   not ishorizontalrule(lines[0]) and \
                   ishorizontalrule(lines[1])):
            lines.pop(0)
        name = lines.pop(0).strip('# ' + os.linesep)
        lines.pop(0)
        variables = []
        while ConfigVariable.isVariable(lines):
            variables.append(ConfigVariable.parse(lines))
        return cls(name, variables)
    
    isSection = staticmethod(isSection)
    parse = classmethod(parse)

class ConfigVariable:
    def __init__(self, name, enabled=False, value="", doc=""):
        self.name = name
        self.enabled = enabled
        self.value = value
        self.doc = doc
    
    def __repr__(self):
        c = self.__class__.__name__
        return "<%s name='%s' %svalue='%s'>" % \
            (c, 
             self.name, 
             self.enabled and "enabled " or "",
             self.value)
    
    def __str__(self):
        retval = ""
        doclines = self.doc.strip(os.linesep).split(os.linesep)
        if len(doclines[0]) > 0:
            for line in self.doc.strip(os.linesep).split(os.linesep):
                retval += "# " + line + os.linesep
        if not self.enabled:
            retval += "#"
        retval += "py[\"" + self.name + "\"] = "
        if isinstance(self.value, str):
            retval += "\"" + self.value + "\"" + os.linesep
        else:
            retval += str(self.value) + os.linesep
        retval += os.linesep
        return retval
    
    def isVariable(lines):
        if len(lines) == 0:
            return False
        emptybefore = isblank(lines[0])
        for l in lines:
            if emptybefore and isblank(l):
                continue
            elif isblank(l):
                break
            emptybefore = False
            if ConfigVariable.varStart(l):
                return True
            elif iscomment(l):
                continue
            else:
                break
        return False
    
    def varStart(line):
        return line.lstrip('# ').startswith("py[")
    
    def parse(cls, lines):
        while isblank(lines[0]):
            lines.pop(0)
        doc = value = ""
        while iscomment(lines[0]) and \
              not ConfigVariable.varStart(lines[0]):
            doc += lines.pop(0).strip('# ')
        enabled = not iscomment(lines[0])
        name = lines[0]
        name = name[name.find('[')+1:name.find(']')]
        name = eval(name)
        value += lines[0][lines[0].find('=')+1:]
        lines.pop(0)
        # fill value until we reach a blank line
        while len(lines) != 0:
            if isblank(lines[0]):
                break
            if ConfigVariable.varStart(lines[0]):
                break
            value += lines.pop(0).strip('#')
        value = eval(value.rstrip(os.linesep))
        return cls(name, enabled, value, doc)
    
    isVariable = staticmethod(isVariable)
    varStart = staticmethod(varStart)
    parse = classmethod(parse)

