# Copyright (c) 2007 Michael "Z" Goddard
# This software is distributed under the MIT license.

"""
Common file operations, such as path manipulation, used by webfront.
"""

import os
import sys

RESOURCE_DIRECTORY = "resources"
TEMPLATE_DIRECTORY = "templates"
SESSION_FILE = ".sessions.py" # needs to end in a none text file extension
# the session file cannot be viewable except maybe to webfront users that are
# logged in.  Otherwise it should not be visable to the general public.  
# Otherwise any could see the valid session id and the username linked with 
# it.  
# 
# appending .py is a solution to this so that even if its permissions allow
# anyone to read as the normal file permissions are, they won't since it is 
# an executable, or so the server thinks.

def findflavourdir():
    """
    Searches the sys.path list for the Pyblosxom folder and if it finds it
    returns that path with Pyblosxom and flavours appended to it.
    """
    for path in filter(lambda p: os.path.isdir(p), sys.path):
        if "Pyblosxom" in os.listdir(path):
            return os.path.join(path, "Pyblosxom", "flavours")

def templatePath():
    """
    Returns the path to the template directory.
    """
    if __file__.count(os.sep):
        return os.path.join(__file__[:__file__.rfind(os.sep)], 
                            TEMPLATE_DIRECTORY)
    else:
        return TEMPLATE_DIRECTORY

def resourcePath():
    """
    Returns the path to the resource directory.
    """
    return os.path.join(__file__[:__file__.rfind(os.sep)], RESOURCE_DIRECTORY)

def sessionPath():
    """
    Returns the path to the session file.
    """
    return os.path.join(temporaryPath(), SESSION_FILE)

def temporaryPath():
    """
    Returns the path to the temporary directory inside of the webfront 
    package.  Also, makes that directory if it doesn't already exist.
    
    The directory is for files and folders needed by the webfront that are
    generated while running (template caches, session file, etc.).
    """
    directory = __file__
    if directory.count(os.sep) == 0:
        directory = ""
    else:
        directory = directory[:directory.rfind(os.sep)]
    path = os.path.join(directory, "temporary")
    if not os.access(path, os.F_OK):
        os.mkdir(path)
    return path

def getTemporaryFile(name, mode='r'):
    """
    Returns a file-like object from the temporary directory.
    """
    filepath = os.path.join(temporaryPath(), name)
    if not os.access(filepath, os.F_OK):
        tempfile = open(filepath, 'w')
        tempfile.write('')
        tempfile.close()
    return open(filepath, mode)

def file_punter(lookIn, request):
    """
    Delivers files.  Mostly borrowed from filekicker.py except where it checks
    for the file.
    http://snarfed.org/space/filekicker.py
    """
    import mimetypes
    import stat
    import types
    
    # chunk size for transferring files
    BLOCK_SIZE = 4096
    
    # populate this with MIME types to replace with text/plain, for in-browser
    # viewing.
    REDUCE_TO_TEXT_PLAIN = ['text/python', 'text/x-python', 'text/lisp',
                            'text/elisp', 'text/x-diff']
    
    config = request.conf
    response = request.getResponse()
    
    # does the file exist?
    filename = request.http['PATH_INFO']
    print >> sys.stderr, filename
    filename = filename[filename.find(os.sep)+1:]
    filename = os.path.join(lookIn, filename)
    if not os.path.isfile(filename):
        return
    
    # add the headers
    type, encoding = mimetypes.guess_type(filename)
    if type:
        if type in REDUCE_TO_TEXT_PLAIN:
            type = 'text/plain'
        response.addHeader('Content-Type', type)
    
    if encoding:
        response.addHeader('Content-Encoding', encoding)
    
    length = os.stat(filename)[stat.ST_SIZE]
    response.addHeader('Content-Length', str(length))
    
    # send the file to the client
    f = file(filename, 'rb', BLOCK_SIZE)
    while True:
        block = f.read(BLOCK_SIZE)
        if not block:
            break
        response.write(block)
    
    f.close()
    return 1
