# Copyright (c) 2007 Michael "Z" Goddard
# This software is distributed under the MIT license.

"""
Keeps track of the user sessions.
"""

import os
import time
import pickle
from StringIO import StringIO

import fileops

LOGIN_TIME = 24 * 3600
A_YEAR = 365.25 * 24 * 3600

class SessionControl(object):
    """
    SessionControl controls a file to allow for concurrent access.
    """
    def __init__(self, sessionfile=fileops.sessionPath()):
        self._file = sessionfile
        self.has_lock = False
        self.__load_sessions()
        self.__clean_up()
    
    def __del__(self):
        # unlock the file at object release if it has the lock
        if self.has_lock:
            self.__unlock_file()
    
    def __initialize(self):
        """
        Writes the initial data (an empty dict) to the session file.
        """
        pickle.dump({}, open(self._file, 'wb'))
    
    def __clean_up(self):
        """
        Cleans old sessions from the sessiontable.
        """
        for sessionid, session in self.sessiontable.iteritems():
            if session["expire"] < time.time():
                del self[sessionid]
    
    def __load_sessions(self):
        lockedbefore = self.has_lock
        if not self.has_lock:
            if not self.__lock_file():
                raise Exception, "Unable to attain lock on file and read it."
        try:
            try:
                sessionfile = open(self._file, 'rb')
                self.sessiontable = pickle.load(sessionfile)
                sessionfile.close()
            except StandardError, value:
                if value.strerror == "Permission denied":
                    raise
                else:
                    self.__initialize()
                    sessionfile = open(self._file, 'rb')
                    self.sessiontable = pickle.load(sessionfile)
                    sessionfile.close()
        finally:
            # we have to release the lock no matter what or else, no other 
            # session controls can load them, including this one.
            if not lockedbefore:
                self.__unlock_file()
    
    def __save_sessions(self):
        """docstring for __save_sessions"""
        lockedbefore = self.has_lock
        if not self.has_lock:
            if not self.__lock_file():
                raise Exception, "Unable to attain lock on file and write it."
        # first test pickling the session table so that if it doesn't pickle
        # the current valid session file is not corrupted.
        testbuffer = StringIO()
        try:
            try:
                pickle.dump(self.sessiontable, testbuffer)
            except Exception, e:
                raise Exception, "The Pickler did not like something in " + \
                    "the sessiontable so the changes were not saved."
            else:
                sessionfile = open(self._file, 'wb')
                pickle.dump(self.sessiontable, sessionfile)
                sessionfile.close()
        finally:
            if not lockedbefore:
                self.__unlock_file()
    
    def __lock_file(self):
        """
        Simple file lock.  Checks for the existence of a directory sharing 
        the name of the file plus a underscore.  If it exists the file is
        already locked.  Otherwise make the directory and set the locked 
        variables.
        """
        if self.has_lock:
            # wait, why lock when we already have it
            return True
        start = time.time()
        increment = 0.0
        while time.time() - start < 5:
            if time.time() - start > increment:
                increment += 0.1
                # attempt to lock
                if os.access(self._file + '_', os.F_OK):
                    # someone else has it, oh well.  wait for the next interal
                    pass
                else:
                    try:
                        os.mkdir(self._file + '_')
                    except OSError, error:
                        if error.strerror == "Permission denied":
                            raise "Permission denied. Unable to lock session file."
                        elif error.strerror == "File exists" or error.errno == 17:
                            continue
                        else:
                            raise
                    self.has_lock = True
                    return True
            time.sleep(0.1) # sleep the process so we are not busy-waiting
        return False
    
    def __unlock_file(self):
        if self.has_lock:
            try:
                os.rmdir(self._file + '_')
            except:
                pass
            self.has_lock = False
    
    def newSession(self, request):
        session = {}
        session["user"] = request.data["user"]
        session["ip"] = request.http["REMOTE_ADDR"]
        session["client"] = request.http["HTTP_USER_AGENT"]
        session["expire"] = time.time() + LOGIN_TIME
        session["id"] = hash("".join([str(x) for x in session.values()]))
        self[session["id"]] = session
        return session
    
    def has_key(self, key):
        """docstring for has_key"""
        return self.sessiontable.has_key(key)
    
    def __getitem__(self, key):
        """docstring for __getitem__"""
        return self.sessiontable[key]
    
    def __delitem__(self, key):
        """
        lock the file.  load the file. remove the session. save the file.
        unlock the file.
        """
        if not self.__lock_file():
            raise Exception, "... unable to acquire file lock"
        try:
            self.__load_sessions()
            del self.sessiontable[key]
            self.__save_sessions()
        finally:
            self.__unlock_file()
    
    def __setitem__(self, key, session):
        """
        first lock the session file and then load the sessions so any other
        changes are kept.  change the table.  save the sessions. unlock the
        file.
        """
        if not self.__lock_file():
            raise Exception, "... unable to acquire file lock"
        try:
            self.__load_sessions()
            self.sessiontable[key] = session
            self.__save_sessions()
        finally:
            self.__unlock_file()
    

