# Copyright (c) 2007 Michael "Z" Goddard
# This software is distributed under the MIT license.

"""
Base functions and classes used by the webfront.
"""

import os
import sys
import cgi
import time
import Cookie

from Pyblosxom.pyblosxom import Request
from Pyblosxom import tools

from template import *
from session import *
from fileops import *

def webfront_edit_parser(filename, request):
    """
    Entry parser code taken from Pyblosxom.pyblosxom.blosxom_entry_parse.
    While we still need to parse the entry, we don't need any formatting since
    we need the original text to be editted.  aka the naked parser.
    """
    entryData = {}
    
    f = open(filename, "r")
    lines = f.readlines()
    f.close()
    
    # the file has nothing in it...  so we're going to return
    # a blank entry data object.
    if len(lines) == 0:
        return { "title": "", "body": "" }
    
    title = lines.pop(0).strip()
    entryData['title'] = title
    
    # absorb meta data lines which begin with a #
    while lines and lines[0].startswith("#"):
        meta = lines.pop(0)
        meta = meta[1:].strip()     # remove the hash
        meta = meta.split(" ", 1)
        entryData[meta[0].strip()] = meta[1].strip()
    
    entryData['body'] = ''.join(lines)
    
    return entryData

def webfront_directory_scan(directoryscan, depthlimit=1, showhidden=False):
    """
    Scans the given directory for directories and records what is in them.
    
    @param directoryscan: a string of the directory to start at
    @type directoryscan: C{string}
    
    @param depthlimit: how deep to scan
    @type directoryscan: C{int}
    
    @param showhidden: whether to count hidden files (that begin with a '.') 
    and directories 
    
    @returns: C{dict}
    """
    # return the parent directory by removing the stuff after the last or next
    # to last separator, otherwise if there is no parent (no separators)
    # return '.' to show that the parent is the current directory
    parent = lambda directory: directory.count(os.sep) > 1 and \
        directory[directory.find(os.sep)+1:directory.rfind(os.sep)] or '.'
    
    # how deep into the filesystem ... counts the separators
    depth = lambda directory: directory.count(os.sep) - 1
    
    # return the filename ... whatever is after the last separator
    filename = lambda path: path[path.rfind(os.sep)+1:]
    
    directories = {}
    
    # just make sure we were given a path to a directory
    if os.path.isdir(directoryscan):
        dirpath = "." + os.sep
        dirlist = os.listdir(directoryscan)
        
        for directory in dirlist:
            if directory.startswith('.') and not showhidden:
                continue
            dirlist[dirlist.index(directory)] = dirpath + directory
        
        # the originating directory needs to be initialized
        directories['.'] = {"files": [], "directories": []}
        
        # since dirlist will be added to and popped we just loop until we have
        # run out
        while len(dirlist):
            if depth(dirlist[0]) <= depthlimit:
                if os.path.isdir(directoryscan + os.sep + dirlist[0]):
                    if depth(dirlist[0]) < depthlimit:
                        key = dirlist[0][dirlist[0].find(os.sep)+1:]
                        directories[key] = {"files": [], "directories": []}
                    
                    # append the directories in this directory to list to look
                    # at
                    dlist = os.listdir(directoryscan + os.sep + dirlist[0])
                    for directory in dlist:
                        if directory.startswith('.') and not showhidden:
                            continue
                        dirlist.append(dirlist[0] + os.sep + directory)
                    
                    # add this directory to its parent's list of directories
                    directories[parent(dirlist[0])]["directories"].append(
                        filename(dirlist[0]))
                else:
                    if depth(dirlist[0]) <= depthlimit:
                        # add the filename to its parent's list of files
                        directories[parent(dirlist[0])]["files"].append(
                            filename(dirlist[0]))
            dirlist.pop(0)
    return directories

def webfront_file_list(directoryscan, extfilter, showhidden=False):
    """
    Used instead of the entry list callback to get files and dates and
    filter those files by extensions.
    """
    
    files = os.listdir(directoryscan)
    # filter out hidden files
    files = [f for f in files if not (f.startswith('.') and not showhidden)]
    # filter out the directories
    files = [f for f in files \
             if os.path.isfile(os.path.join(directoryscan, f))]
    # if we were given an extension filter, use it to weed out what is not 
    # wanted
    if extfilter != None:
        files = [f for f in files if f[f.rfind(os.extsep)+1:] in extfilter]
    # get mtime
    # also usable for later additions
    files = [{'name': f, \
        'mtime': os.stat(os.path.join(directoryscan, f)).st_mtime, \
        'ext': f[f.rfind(os.extsep)+1:]} \
        for f in files]
    files.sort(lambda x, y: 
        x['mtime'] > y['mtime'] and -1 or \
        x['mtime'] == y['mtime'] and 0 or 1)
    
    return files

def resource_punter(request):
    return file_punter(resourcePath(), request)

def save_changes(filename, title, body, metadata={}):
    savefile = open(filename, 'w')
    title = title.strip()
    if title.rfind(os.linesep) != -1:
        savefile.write(title)
    else:
        savefile.write(title + os.linesep)
    for key, data in metadata.items():
        savefile.write("# %s %s%s" % (key, data, os.linesep))
    savefile.write(os.linesep) # a line to split the meta data from other lines
    if body:
        savefile.write(body.strip())
    savefile.close()

def save_draft(filename, title, body, metadata={}):
    draftname = filename[:filename.rfind(os.extsep)+1] + "draft"
    metadata['webfront:extension'] = filename[filename.rfind(os.extsep)+1:]
    save_changes(draftname, title, body, metadata)

def upload_file(filepath, fileobj):
    if not fileobj:
        return False
    fout = file (filepath, 'wb')
    while 1:
        chunk = fileobj.read(100000)
        if not chunk: break
        fout.write (chunk)
    fout.close()
    return True

class Webfront(object):
    """
    Manages webfront data and handles requests to the webfront.
    """
    def __init__(self, request, render=True):
        """docstring for __init__"""
        self.request = request
        data = request.data
        self.session = None
        self.template = None
        self._result = None
        self.toRender = render
        self.renderqueue = {}
        self.validation_failed = False
        self.existence_failed = []
        self.handled = 0
        self.sessioncontrol = SessionControl(sessionPath())
        self.pageargs = self.http.get('QUERY_STRING') != None and \
                        cgi.parse_qs(self.http.get('QUERY_STRING')) or \
                        None
        if self.pageargs != None:
            for arg, val in zip(self.pageargs, self.pageargs.values()):
                self.pageargs[arg] = val[0]
            if self.pageargs.has_key('mode'):
                self.mode = self.pageargs['mode']
                data['mode'] = self.mode
                del self.pageargs['mode']
        self.pages = {}
        self.links = []
        tools.run_callback("webfront_page",
            {'pages': self.pages,
             'pagelinks': self.links,
             'webfront': self})
        self.links.append(("Logout", "?edit=logout"))
        self.errors = []
        self.flashes = []
    
    def __del__(self):
        self.saveSession()
    
    def __getattr__(self, name):
        rendername = name[:name.rfind('page')]
        if name.endswith('page') and rendername in self.pages:
            self.action = rendername
            self.template = self.pages[rendername]
            return lambda *args, **kwargs: \
                self.pages[rendername](self, *args, **kwargs)
        elif name == "data":
            return self.request.data
        elif name == "http":
            return self.request.http
        elif name == "form":
            return self.request.getForm()
        elif name in ["config", "conf", "configuration"]:
            return self.request.config
        elif name == "mode":
            return "post"
        elif name == "basedir":
            return self.config['datadir']
        elif name == "pathdir" or name == "pathdirectory":
            return os.path.join(self.basedir, self.pathinfo)
        elif name == "pathinfo":
            return self.http['PATH_INFO'].strip(os.sep)
        raise AttributeError, name
    
    def handle(self):
        """
        The handler function of the webfront.
        """
        self.handled = 1
        request = self.request
        config = request.conf
        data = self.request.data
        http = self.request.http
        
        data["webfront_action"] = self.function
        # add the draft page extension so that they are picked up
        data['extensions']['draft'] = webfront_edit_parser 
        data['currentdir'] = config['base_url'] + os.sep + http['PATH_INFO'].lstrip(os.sep)
        while not os.path.isdir(self.pathdir):
            currdir = data['currentdir']
            data['currentdir'] = currdir[:currdir.rfind(os.sep)]
            if len(data['currentdir']) < len(config['base_url']):
                data['currentdir'] = config['base_url']
                break
        
        # is the client logged in?
        if not self.isLoggedIn():
            # can we log the client in?
            if not self.logUserIn():
                # let the client log in by displaying the form
                self.loginpage()
                self.render()
                return
        # the following is code to let future interactions know what was 
        # previously done by webfront. this may not be necessary, but here
        # in case needed by later code.  REVALULATE LATER
        if self.session.has_key("actions"):
            self.session["actions"].append({'function': self.function})
        else:
            self.session["actions"] = [{'function': self.function}]
        # try:
        args = dict(self.pageargs)
        del args['edit']
        self.renderpage(self.function, **args)
        if self.toRender:
            self.render()
        # except Exception, e:
        #    self.result = False
        #    self.renderpage('entry')
        #    self.error("""Oops, an unexcepted error occured and we're not sure how to handle it.""", str(e))
        #    if self.toRender:
        #         self.render()
        return self.handled
    
    def renderpage(self, rendername="entry", *args, **kwargs):
        request = self.request
        data = request.data
        config = request.conf
        http = request.http
        if rendername == "logout":
            self.logUserOut()
            # replace the following with another page like the blog itself,
            # maybe the home page
            data["webfront_action"] = "entry"
            self.loginpage()
            return
        elif rendername in self.pages:
            self.action = rendername
            return self.pages[rendername](self, *args, **kwargs)
        else:
            self.error("The given action %s does not exist or has not been implemented." % rendername)
            self.renderpage()
    
    def render(self, subname=None, queueup=False, *args, **kwargs):
        data = self.request.data
        if subname:
            if self.template and not queueup:
                self.template.render(subname, *args, **kwargs)
            else:
                if self.renderqueue.has_key(subname):
                    self.renderqueue[subname].append(data.copy())
                else:
                    self.renderqueue[subname] = [data.copy()]
            return
        # if not self.template:
        #     raise TemplateException, "Cannot render template without one assigned"
        savedata = data
        for sub, subdata in self.renderqueue.items():
            for sd in subdata:
                data.update(sd)
                self.template.render(sub)
        data.update(savedata)
        self.template.render()
    
    def saveSession(self):
        if self.session:
            self.sessioncontrol[self.session["id"]] = self.session
    
    def error(self, text, details=None):
        data = self.request.data
        data['error_text'] = text
        if details != None:
            data['error_detail'] = details
            if self.pageargs.has_key('errordetails') and self.pageargs['errordetails'][0].lower() == "yes" and 'error_showdetail' in self.template:
                    self.render('error_showdetail')
            elif 'error_hidedetail' in self.template:
                self.render('error_hidedetail')
        elif data.has_key('error_detail'):
            del data['error_detail']
        self.render('error', queueup=True)
        self.errors.append((text, details))
    
    def flash(self, text):
        self.data["flash_text"] = text
        self.render("flash")
        self.flashes.append(text)
    
    def getResult(self):
        form = self.form
        if self._result == None:
            return form.has_key('result') and [True] + form.getlist('result') or [False]
        else:
            return form.has_key('result') and [self._result] + form.getlist('result') or [self._result]
    
    def setResult(self, value):
        self._result = value
    
    def getMode(self):
        if hasattr(self, '_mode'):
            return self._mode
        else:
            return 'post'
    
    def setMode(self, mode):
        self._mode = mode
        if mode == 'post':
            self.extfilter = 'extensions'
            self.basedirectory = 'datadir'
        elif mode == 'file':
            self.extfilter = None
            if self.config.has_key('webfront_filedir'):
                self.basedirectory = 'webfront_filedir'
            else:
                self.basedirectory = 'datadir'
            self.textextensions = []
        elif mode == 'flav':
            self.extfilter = None
            self.basedirectory = findflavourdir()
            self.textextensions = []
        else:
            self.setMode('post')
    
    def getBaseDirectory(self):
        if hasattr(self, '_basedirectory'):
            return self._basedirectory
        else:
            return self.config['datadir']
    
    def setBaseDirectory(self, directory):
        if self.config.has_key(directory):
            self._basedirectory = self.config[directory]
        else:
            self._basedirectory = directory
        if not os.path.isdir(self._basedirectory):
            # oops, defaulting to datadir
            self.flash("It seems that %s is not a directory so we are displaying the data directory." % directory)
            self._basedirectory = self.config['datadir']
    
    def getExtFilter(self):
        if hasattr(self, '_extfilter'):
            return self._extfilter
        else:
            return self.data['extensions']
    
    def setExtFilter(self, filefilter):
        self._extfilter = []
        if self.data.has_key(filefilter) and type(self.data[filefilter]) is type(dict()):
            self._extfilter = self.data[filefilter]
        elif type(filefilter) is type(str()) and type(eval(filefilter)) is type(list()):
            evalfilter = eval(filefilter)
            for ext in evalfilter:
                if self.data['extensions'].has_key(ext):
                    self._extfilter[ext] = self.data['extensions'][ext]
                else:
                    self._extfilter[ext] = None
        elif type(filefilter) is type(str()) and type(eval(filefilter)) is type(dict()):
            self._extfilter = eval(filefilter)
        elif filefilter == None or type(filefilter) is type(str()) and eval(filefilter) == None:
            self._extfilter = None
    
    def getTextExtensions(self):
        if hasattr(self, '_textextensions'):
            return self._textextensions
        else:
            return []
    
    def setTextExtensions(self, extensions):
        extensions += ["txt", "html", "htm", "css", "js"]
        if self.config.has_key('self_textextensions'):
            extensions.append(self.config['self_textextensions'])
        self._textextensions = extensions
    
    def getFunction(self):
        return self.pageargs['edit']
    
    def inWebfront(request):
        if isinstance(request, Webfront):
            webfront = request
            if webfront.pageargs.has_key('edit'):
                return True
            request = request.request
        if not isinstance(request, Request):
            raise TypeError, "need Request, %s found" % request.__class__.__name__
        pyhttp = request.http
        qstr = pyhttp.get('QUERY_STRING')
        if qstr != None:
            parsed_qs = cgi.parse_qs(qstr)
            return parsed_qs.has_key('edit')
        return False
    
    def replacecheck(self, path, ignoreext=True):
        """
        Checks the given path to see if the containing folder contains a
        file sharing the same name, ifnoring whether the extension is 
        different or not. If such a file exists it'll return the file name. 
        If a file doesn't exist it'll do nothing and return 
        'not replacing'. If the page has already been shown and an answer 
        selected it will react to that responding with a boolean value 
        depending on the button selected.
        
        @param path: filesystem path to check if a file already exists
        @type path: string
        
        @returns: string, 'not replacing', True, or False
        """
        if isinstance(path, list):
            for p in path:
                retval = self.replacecheck(p, ignoreext)
                if isinstance(retval, bool):
                    return retval
                elif retval != "not replacing":
                    return "replacing"
            return "not replacing"
        
        # check for the replacepage first
        if self.form.has_key('replace'):
            if self.form.getfirst('replace').lower() == "yes":
                return True
            elif self.form.getfirst('replace').lower() == "no":
                return False
            else:
                # hmm, while replace exists, neither of the choices are selected
                raise MalformedTemplateException, """
                The form of your replace template requires a input with the
                name of 'replace' with a value of either 'Yes' or 'No'."""
        
        # get the names of the files in the directory
        files = os.listdir(path[:path.rfind(os.sep)])
        
        filenameext = path[path.rfind(os.sep)+1:]
        filename, fileext = os.path.splitext(filenameext)
        for f in files:
            # check the names of the file in the directory and that of the
            # file given by the parameter path, ignoring the extension.
            if self.mode == 'post' and \
               (fileext == ".draft" and \
               os.path.splitext(f)[-1] != ".draft") or \
               (fileext != ".draft" and \
               os.path.splitext(f)[-1] == ".draft"):
                # we give special treatment to drafts while in post mode. If
                # we are testing a draft we will only check against other 
                # drafts.  If it isn't a draft we are checking, leave drafts
                # alone.
                continue
            if ignoreext and os.path.splitext(f)[0] == filename:
                return f
            elif not ignoreext and f == filenameext:
                return f
        return 'not replacing'
    
    def getfield(self, field, default=None):
        form = self.form
        action = getattr(self, "action", "webfront") + "_"
        return form.getfirst(action + field, default)
    
    def setfield(self, field, value):
        form = self.form
        data = self.request.data
        action = getattr(self, "action", "webfront") + "_"
        data[action + field] = value
    
    def renewfield(self, field, default=None):
        self.setfield(field, self.getfield(field, default))
    
    def validate(self, field, test, response):
        if test == 'existence':
            if self.getfield(field) == None:
                if isinstance(response, list):
                    self.error(*response)
                else:
                    self.error(response)
                self.validation_failed = True
                self.existence_failed.append(field)
            return
        elif not field in self.existence_failed and test(self.getfield(field)):
            if isinstance(response, list):
                self.error(*response)
            else:
                self.error(response)
            self.validation_failed = True
    
    def isLoggedIn(self):
        # check the cookies for the user
        request = self.request
        config = request.conf
        data = request.data
        http = request.http
        cookies = Cookie.SimpleCookie( request.http.get("HTTP_COOKIE", {}) )
        if cookies.has_key("userid"):
            data["id"] = int(cookies["userid"].value)
            if not self.sessioncontrol.has_key(data["id"]):
                return False
            self.session = self.sessioncontrol[data["id"]]
            data["session"] = self.session
            # check client and ip address
            if not http["HTTP_USER_AGENT"] == data["session"]["client"] and \
               not http["REMOTE_ADDR"] == data["session"]["ip"]:
                return False
            data["user"] = self.session["user"]
            if not config.has_key('webfront_users') or \
               data['user'] not in config['webfront_users']:
                return False
            # logged in, so lets update the expire times
            data["session"]["expire"] = time.time() + LOGIN_TIME
            self.sessioncontrol[data["session"]["id"]] = data["session"]
            expiretime = time.strftime("%a %d-%b-%Y %H:%m:%S GMT", time.gmtime(data["session"]["expire"] + A_YEAR))
            pyblosxompath = config["base_url"]
            pyblosxompath = pyblosxompath[pyblosxompath.find(request.http["HTTP_HOST"])+len(request.http["HTTP_HOST"]):]
            request.getResponse().addHeader("Set-Cookie", "userid=%s; expires=%s; path=%s" % (data["session"]["id"], expiretime, pyblosxompath))
            return True
    
    def hasLoggedIn(request):
        cookies = Cookie.SimpleCookie( request.http.get("HTTP_COOKIE", {}) )
        return cookies.has_key("userid")
    
    def logUserIn(self, user=None, pwd=None):
        request = self.request
        form = self.form
        if form.getfirst("name") and form.getfirst("pass") or user and pwd:
            data = request.data
            # grab all the users from the config file
            validUsers = request.conf.get("webfront_users", {})
            
            if len(validUsers) == 0:
                self.error("You have not defined any usernames yet.")
                return False
            
            if not user:
                user = form.getfirst("name")
                pwd  = form.getfirst("pass")
            if not user in validUsers or ( user in validUsers and validUsers[user] != pwd ):
                self.error("The username and password given do not match any accounts.")
                return False
            
            response = request.getResponse()
            data["user"] = form.getfirst("name")
            self.session = self.sessioncontrol.newSession(request)
            data["session"] = self.session
            userid = data["session"]["id"]
            expiretime = time.strftime("%a %d-%b-%Y %H:%m:%S GMT", time.gmtime(time.time()+A_YEAR))
            pyblosxompath = request.conf["base_url"]
            pyblosxompath = pyblosxompath[pyblosxompath.find(request.http["HTTP_HOST"])+len(request.http["HTTP_HOST"]):]
            response.addHeader("Set-Cookie", "userid=%s; expires=%s; path=%s" % (userid, expiretime, pyblosxompath))
            return True
        elif form.getfirst("name") or form.getfirst("pass"):
            self.error("You must enter in both a username and password.")
            return False
    
    def logUserOut(self):
        request = self.request
        data = request.data
        userid = data["session"]["id"]
        del self.sessioncontrol[userid]
        response = request.getResponse()
        expiretime = time.strftime("%a %d-%b-%Y %H:%m:%S GMT", time.gmtime(time.time() + A_YEAR))
        pyblosxompath = request.conf["base_url"]
        pyblosxompath = pyblosxompath[pyblosxompath.find(request.http["HTTP_HOST"])+len(request.http["HTTP_HOST"]):]
        response.addHeader("Set-Cookie", "userid=%s; expires=%s; path=%s" % (hash(time.time()), expiretime, pyblosxompath))
    
    inWebfront = staticmethod(inWebfront)
    hasLoggedIn = staticmethod(hasLoggedIn)
    function = property(getFunction)
    result = property(getResult, setResult)
    mode = property(getMode, setMode)
    basedirectory = property(getBaseDirectory, setBaseDirectory)
    basedir = basedirectory
    extfilter = property(getExtFilter, setExtFilter)
    textextensions = property(getTextExtensions, setTextExtensions)

class WebfrontMode(object):
    """
    Currently empty but will be populated with the webfronts methods for 
    modes.
    """
    pass

class WebfrontUser(object):
    """
    Methods and variables involving a user of the webfront.
    """
    pass

class WebfrontAjax(Webfront):
    """
    This class handles ajax requests.
    
    This could do stuff like when rendering it may only render the body area
    of a template or maybe a smaller section by request.  So that the server
    side does not have to generate the full page and have it wasted on the
    client side.
    """
    pass


def register_configuration():
    from configedit import ConfigEditor, ConfigSection, ConfigVariable
    def dewhite(string):
        string = string.splitlines()
        for line, index in zip(string, range(len(string))):
            string[index] = line.strip()
        string = ''.join(string)
        return string
    
    webfront_config = ConfigSection(
        name="Webfront Configuration",
        variables=[
            ConfigVariable(
                name="webfront_users",
                value={'username': 'password'},
                enabled=True,
                doc=dewhite("""
                Defines what usernames with accompanying passwords may be 
                allowed to log into the webfront and edit posts, flavours, 
                and files.""")
            ),
            ConfigVariable(
                name="webfront_link",
                value="<a href='%s'>Webfront</a>",
                enabled=False,
                doc=dewhite("""
                Defines the text of the link to the webfront placed in a 
                flavour by the variable $webfrontlink.  %s must appear once, 
                this is where the link's address is placed.
                """)
            ),
            ConfigVariable(
                name="webfront_editlink",
                value="<a href='%s'>Edit</a>",
                enabled=False,
                doc=dewhite("""
                Defines the text of the link to the edit page of the post 
                that it appears by as the flavour variable $editlink in the
                story template.  %s must appear once in the string, this is
                where the link's address is placed.
                """)
            ),
            ConfigVariable(
                name="webfront_filedir",
                value="/path/to/desired/directory",
                enabled=False,
                doc=dewhite("""
                The file mode of the webfront can be directed to any directory
                the web server has access to.  By default the file mode reads
                from the datadir
                """)
            )
        ]
    )
    ConfigEditor.registerSection(webfront_config)

def test_configuration(config):
    """
    Tests to see that the webfront is properly configured and if not return a
    string detailing what is missing.
    
    Incomplete.
    
    @returns: None or str
    """
    error_string = ""
    
    # test for webfront_users for login
    if not config.has_key('webfront_users') or \
       not isinstance(config['webfront_users'], dict) or \
       len(config['webfront_users']) == 0:
        error_string += """
        <p>The configuration variable 'webfront_users' either does not exist or
        has been set incorrectly, please check to make sure that it is a
        dictionary with pairs that equate to your desired username, password
        pairs.</p>
        
        For example for a username of <i>Joe</i> and a password of 
        <i>joeisawesome</i>, set 'webfront_users' to {'Joe': "joeisawesome"}."""
    
    if len(error_string):
        return error_string

